<Query Kind="Program">
  <NuGetReference>Microsoft.AspNetCore.Http.Abstractions</NuGetReference>
  <NuGetReference>Microsoft.AspNetCore.Http.Extensions</NuGetReference>
  <NuGetReference>Swashbuckle.AspNetCore.Swagger</NuGetReference>
  <NuGetReference>Swashbuckle.AspNetCore.SwaggerGen</NuGetReference>
  <NuGetReference>Swashbuckle.AspNetCore.SwaggerUI</NuGetReference>
  <NuGetReference>Unity.Microsoft.DependencyInjection</NuGetReference>
  <Namespace>LINQPad.Controls</Namespace>
  <Namespace>Microsoft.AspNetCore.Builder</Namespace>
  <Namespace>Microsoft.AspNetCore.Hosting</Namespace>
  <Namespace>Microsoft.AspNetCore.Http</Namespace>
  <Namespace>Microsoft.AspNetCore.Mvc</Namespace>
  <Namespace>Microsoft.Extensions.DependencyInjection</Namespace>
  <Namespace>Microsoft.Extensions.FileProviders</Namespace>
  <Namespace>Microsoft.Extensions.Hosting</Namespace>
  <Namespace>Microsoft.Extensions.Logging</Namespace>
  <Namespace>System.Text.Json</Namespace>
  <Namespace>System.Threading.Tasks</Namespace>
  <Namespace>Unity</Namespace>
  <Namespace>Unity.Microsoft.DependencyInjection</Namespace>
  <IncludeAspNet>true</IncludeAspNet>
  <IncludeWinSDK>true</IncludeWinSDK>
</Query>

void Main()	// Minimal services
{
	WebApplication.CreateBuilder([])
		.ConfigureApp()
		.Run();
}
//::::::::::[ PAGES ]:::::::::://
public class HomePage : HtmlBase {
	public HomePage(WeatherForecastController weather, IUnityContainer container, IFoo info) {
		// We have to use Unity to resolve IWeatherFormatter since it has a [Dependency]
		// IFooBar Bar {get;set;} property - only Unity supports "setter" injection"
		var formatter = container.Resolve<IWeatherFormatter>();
		var formatterBarMsg = formatter.Bar.GetMessage(); // Setter injection injects Bar
		var tableData = formatter.Format(weather.Get());  // Get table data and format it
		var sysIFooMsg = info.GetMessage();	// We inject IFoo from system IOC
		
		// Content to be displayed in Body of HTML
		bodyHtml =$"<i>{DateTime.Now}</i> | <b>Unity IFoo</b> says: &quot{formatterBarMsg}&quot"+
				  $" | <b>System IFoo</b> says: &quot;{sysIFooMsg}&quot;" +
				  $"<hr>{tableData}"; 
	} 
}
public class WeatherPage : HtmlBase {
	public WeatherPage(WeatherForecastController weather, IWeatherFormatter formatter) {
		var tableData = formatter.Format(weather.Get()); // Get table data and format it

		// System IOC does not support setter injection so formatter.Bar will always be null
		// To demonstrate this will do a null check on formatter.Bar (which is set in HomePage)
		var formatterStatus = formatter.Bar == null ? "Formatter.Bar is &lt;null&gt;" : "NA";

		// Content to be displayed in Body of HTML
		bodyHtml = $"<p><i>{DateTime.Now}</i> | <b>{formatterStatus}</b></p>" +
				   $"<hr>{tableData}";
	}
}
public class ToggleService : HtmlBase { // Javascript (HtmlBase) request only (toogles IsJson)
	public ToggleService() {
		// WebAppBuilderExtension is the primary component that takes all the
		// decoupled components and registers them in the container.  It has
		// factory logic that will determine whether to use Json or Table formatter
		// for the IWeatherFormatter - injected into HomePage and WeatherPage
		WebAppBuilderExtension.IsJson = !WebAppBuilderExtension.IsJson;
		
		// The service is being invoked by javascript - it doesn't know C#
		// so after we toggle the C# IsJson we'll return its value 
		bodyHtml = JsonSerializer.Serialize(WebAppBuilderExtension.IsJson);
 	}
}
// -----
public class HtmlBase
{
	// So we can see behaviors when Transient registrations are changed to Singleton
	public static int ClickCount { get; set; }
	
	// Shown in the <Body> of the Html below
	public string bodyHtml { get; set; }
	
	// Convenience propery to GetHtml()
	public string Message => GetHtml();
	
	// invoked by Message readonly property
	public string GetHtml()	{
		var isJsonState = WebAppBuilderExtension.IsJson ? "Forecast=json" : "Forecast=table";
		var pageName = GetType().Name.Replace("Page", ""); // HomePage and WeatherPage  
		var completeHtml = $@"<!doctype html>
		<html>
		  <head>
		    <title>{pageName}</title>
		  </head>
		  <body style='background-color:white'>
				<!-- Home and Weather links along with clock and click count -->
	 			<a href='/'>Home</a>&nbsp;|&nbsp;<a href='/Weather'>Weather</a>&nbsp;|&nbsp;
				<span id='clock'></span> Click count: {ClickCount++}
				<h2>{pageName}</h2>  <!-- Display the page name in H2              -->
 			  	<p>{bodyHtml}</p>    <!-- Boddy content provided by parent classes -->

				<!-- If the button is clicked then toggle the Json/Table state     -->
				<button onclick='invokeTogglePage()'>Toggle forecast format</button>
				<span id='buttonInfo'>{isJsonState}</span>
				<hr>
		  </body>
		  <script> 
		  	// Get clock and button info element instances
			var clockElement = document.getElementById( 'clock' );   
			var buttonInfo = document.getElementById( 'buttonInfo' );
			// Set the clock to the current time - initialize
			clock.innerHTML = new Date().toLocaleTimeString(); 
			// Update the clock with time when invoked (called by setInterval below)
			function updateClock( clock ) {{ 
				clock.innerHTML = new Date().toLocaleTimeString();
			}}
			// Invoked by the button click - in html body
			function invokeTogglePage() {{
				// Invoke the TogglePage service and return the current state (true/false) 
				fetch(""http://localhost:5000/TogglePage"") 
				.then(response => {{
					if (!response.ok) {{
						throw new Error('Network response was not ok ' + response.statusText);
					}}
					return response.json();
				}})
				.then(data => {{ 
					// update the buttonInfo element with the data retrieved from serviceq
					buttonInfo.innerHTML = data ? 'Forecast=json' : 'Forecast=table';
				}})
				.catch(error => console.error('Problem with your fetch operation:', error));
			}}
			// Every second update the clockElement with the current time 
			setInterval(function() {{updateClock( clockElement ); }} , 1000);
		</script>
		</html>";
		return completeHtml;
	}
}

//:::::::::::::[ FORMATTERS ]:::::::::::::/
public class TableFormatter : IWeatherFormatter
{
	// Setter injection - [Dependency] is a Unity Attribute
	[Dependency] public IFoo Bar { get; set; } // Only set if Unity Resolves
	
	// Format retrieved weather forecast list as HTML
	public string Format(IEnumerable<WeatherForecast> forecastList)
	{
		var rowData = "";
		Func<object, string> f = (celldata) => { return $"<td>{celldata}</td>"; };
		foreach (var fc in forecastList)
			rowData += $"<tr>{f(fc.Date)}{f(fc.Summary)}{f(fc.TempC)}{f(fc.TempF)}</tr>";
		return $"<table width=500px border=1>{rowData}</table>";
	}
}
public class JsonFormatter : IWeatherFormatter
{
	// Setter injection - [Dependency] is a Unity Attribute
	[Dependency] public IFoo Bar { get; set; } // Only set if Unity Resolves

	// Serialize the retrieved weather forecast list as json
	public string Format(IEnumerable<WeatherForecast> forecastList)
	{
		return JsonSerializer.Serialize(forecastList);
	}
}
//-----
public interface IWeatherFormatter { 
	IFoo Bar {get;set;}
	string Format(IEnumerable<WeatherForecast> forecastList); 
}

//:::::::::[ WEBAPP BOOTSTRAP ]:::::::::://
public static class WebAppBuilderExtension {
	public static bool IsJson {get;set;} = true;  // State for Json/Table formatters
	public static WebApplication ConfigureApp(this WebApplicationBuilder builder) {
		// Display a iFrame so we can launch our browser in LinqPad
		new IFrame(new Uri("http://localhost:5000"), "850px", "320px").Dump();
		var container = new UnityContainer();
		builder.Services.AddTransient<WeatherForecastController>();

		// Factory pattern required for IWeatherFormatter since the
		// Formatter used (Json or Table) will be determined by state
		Func<IServiceProvider, IWeatherFormatter> sysfmtFactory = (provider) => {
			return IsJson 	? provider.GetService<JsonFormatter>()
							: provider.GetService<TableFormatter>();	};
		Func<IUnityContainer, IWeatherFormatter> unityfmtFactory = (provider) => {
			return IsJson	? provider.Resolve<JsonFormatter>()
							: provider.Resolve<TableFormatter>();		};
		// ServiceCollection is container agnostic (although more aligned
		// to the system container, particularly with factories (above)
		var serviceProvider = new ServiceCollection()
			.AddTransient<IWeatherFormatter>(sysfmtFactory)
			.AddSingleton<IUnityContainer>(container)
			.AddTransient<HomePage>()
			.AddTransient<WeatherPage>()
			.AddTransient<ToggleService>()
			.AddTransient<IFoo,NotBar>()
			// Unlike Unity, the system can only resolve registered types
			.AddTransient<JsonFormatter>()  // Not required for unity
			.AddTransient<TableFormatter>() // Not required for unity
		;
		// iterate through our service container and register types with
		// both the system IOC and Unity Container
		foreach (var service in serviceProvider){
			// Add to system IOC (no special logic)
			builder.Services.Add(service); 
			// Add to Unity container - We only have one factory so we can do this
			if (service.ImplementationFactory != null)	
				container.RegisterFactory(service.ServiceType,unityfmtFactory);
			else
				container.RegisterType(service.ServiceType, service.ImplementationType);
		}
		var app = builder.Build();

		// This is what is nice about Unity container - it let's you register types
		// AFTER the builder is built.  For system IOC, once it is built you cannot
		// register any more types (without expensive rebuild)
		//
		// Last come, last serve.  Above both the System and Unity containers had
		// AddTransient<IFoo, NotBar>() - this will replace IFoo in Unity container
		container.RegisterType<IFoo,Bar>();

		// Injected information will be sent to console 
		app.UseMiddleware<MyCustomMiddleware>();
		
		// Map our routes - HomePage, WeatherPage, and ToggleService will be injected
		// by system IOC with its constructor parameters
		app.MapGet("/", (HomePage home) => Results.Content(home.Message, "text/html"));
		app.MapGet("/Weather", (WeatherPage info) => Results.Content(info.Message, "text/html"));
		app.MapGet("/TogglePage", (ToggleService toggle) => Results.Content(toggle.bodyHtml));
		return app;
	}
}

//::::::::: WeatherForecast classes ::::::::::://
public class WeatherForecastController : ControllerBase
{
	private static readonly string[] Summaries = new[]
	{
		"Freezing", "Bracing", "Chilly", "Cool", "Mild", 
		"Warm", "Balmy", "Hot", "Sweltering", "Scorching"
	};

	private readonly ILogger<WeatherForecastController> _logger;
	public WeatherForecastController(ILogger<WeatherForecastController> logger)
	{
		_logger = logger;
	}
	public IEnumerable<WeatherForecast> Get()
	{
		return Enumerable.Range(1, 5).Select(index => new WeatherForecast
		{
			Date = DateOnly.FromDateTime(DateTime.Now.AddDays(index)),
			TempC = Random.Shared.Next(-20, 55),
			Summary = Summaries[Random.Shared.Next(Summaries.Length)]
		})
		.ToArray();
	}
}
public class WeatherForecast
{
	public DateOnly Date { get; set; }
	public int TempC { get; set; }
	public int TempF => 32 + (int)(TempC / 0.5556);
	public string Summary { get; set; }
}

//:::::::::[ FOO ]:::::::::::://
public interface IFoo
{
	string GetMessage();
}
public class NotBar : IFoo
{
	public string GetMessage() { return "Hello World"; }
}
public class Bar : IFoo
{
	[Dependency] public MsgHolder Holder;
	public string GetMessage()
	{
		return Holder.Message;
	}
}
public class MsgHolder
{
	public string Message { get; set; } = "This is FooBar";
}

//::::::::::[ MIDDLEWARE ]:::::::://
public class MyCustomMiddleware
{
	private readonly RequestDelegate _next;

	public MyCustomMiddleware(RequestDelegate next)
	{
		_next = next;
	}

	// IUnityContainer and IFoo is injected into InvokeAsync
	public async Task InvokeAsync(HttpContext httpContext, IUnityContainer container, IFoo foo)
	{
		var unitySays = container.Resolve<IFoo>().GetMessage();
		var systemSays = foo.GetMessage();
		Console.WriteLine($":: Unity Says: {unitySays}\r\n:: System says: {systemSays}");
		
		await _next(httpContext);
	}
}
